@@ -1,10 +1,12 @@
 import asyncio
 import aiohttp
+from collections import Counter
 from dataclasses import dataclass
 import ipaddress
 import logging
 import platformdirs
 import random
+import re
 import socket
 import os
 from sqlmodel import Session, SQLModel, create_engine, select
@@ -37,7 +39,8 @@ def cli(return_help_text=False):
         "--minimum-match",
         type=int,
         default=2,
-        help="an IP address is considered valid after this number of idential responses (default: 2)",
+        help="an IP address is considered valid after this number of idential responses"
+        + " (default: 2)",
     )
     parser.add_argument(
         "--overkill",
@@ -45,6 +48,14 @@ def cli(return_help_text=False):
         default=0,
         help="number of initial requests to make beyond minimum-match (default: 0)",
     )
+    parser.add_argument(
+        "--majority-ratio",
+        type=int,
+        default=3,
+        help="minimum ratio needed to overrule a conflicting response; must be an integer;"
+        + " a value of 2 means 2:1, or that 6 responses of 8.7.8.7 are needed to overrule"
+        + " 3 responses of 7.8.4.4 (default: 3)",
+    )
     parser.add_argument(
         "--max-failures",
         type=int,
@@ -131,6 +142,7 @@ class MyoriginArgs:  # see '--help' for descriptions
     timeout: int = 12000
     minimum_match: int = 2
     overkill: int = 0
+    majority_ratio: int = 3
     max_failures: int = 10
     max_connections: int = 10
     dbfile: str = ''  # ''==db_pathname(), '-'==none, other==alternate file
@@ -199,7 +211,15 @@ def http_timer():  # configure signals to record beginning and end of http reque
     return trace_config
 
 
-async def http_get(url: str, session: aiohttp.ClientSession, max_size=-1, ip_version=0):
+async def http_get(url: str, session: aiohttp.ClientSession, max_size=-1):
+    if hasattr(http_get, 'pytest'):  # mock responses; see myorigin_test.py
+        if http_get.pytest == 1:
+            try:
+                return re.match(r'http.*/(.*)', url).group(1), 290  # whatever follows final slash
+            except:
+                return f'test html response with no IP address', 180
+        elif http_get.pytest == 2:
+            raise ValueError("404 pytest simulated error")
     timer_marks = list()
     # change user agent string for better site compatibility and anonymity, though one site
     # ... suggests a Firefox user agent (https://www.whatismyip.com/automation-rules/);
@@ -286,12 +306,55 @@ def weighted_random(options: dict):
             return k
         total -= weight
     assert False, f"bug in weighted_random, options {options}"
-    # #test:
-    # o = {200: 5, 202: 10, 203: 11, 210: 1, 214: 10}
-    # n = 1000  # number of iterations
-    # h = [k for k in [weighted_random(o) for i in range(n*sum(o.values()))]]
-    # r = dict(sorted({i:h.count(i)/n for i in h}.items()))
-    # #visually compare o and r
+
+
+def are_we_there_yet(completed_jobs, args, logger):
+    top_ip_count = 0  # max of IPv4 and IPv6
+    top_ip_target = args.minimum_match
+    for v in (4, 6):
+        fail_count = sum(1 for r in completed_jobs if r.rtt == 0)
+        if fail_count >= args.max_failures:
+            raise NetworkError(f"{fail_count} requests failed; giving up")
+        votes = Counter([r.ip for r in completed_jobs if (r.ip_version == v and r.rtt != 0)])
+        if len(votes) == 0:
+            continue
+        ranked = sorted(votes.items(), key=lambda i: i[1], reverse=True)
+        # ranked[0][0] is #1 candidate IP; ranked[0][1] is its vote count
+        assert sum(1 for r in completed_jobs if (r.rtt != 0 and r.ip == '')) == 0
+        if len(ranked) > 1 and not hasattr(are_we_there_yet, 'found_multiple_ips'):
+            are_we_there_yet.found_multiple_ips = True  # only show once
+            logger.info(f"multiple IPs received: {[k[0] for k in ranked]}")
+        overrule = 0 if len(ranked) <= 1 else ranked[1][1] * args.majority_ratio
+        target = max(args.minimum_match, overrule)  # votes necessary to win
+        if ranked[0][1] >= target:  # #1 candidate has enough votes
+            ip = ranked[0][0]
+            msg = f"IP found: {ip} ({ranked[0][1]} successes,"
+            logger.info(f"{msg} {fail_count} failures)")
+            if len(ranked) > 1:  # retroactively mark other candidates as failed
+                for r in completed_jobs:
+                    if r.ip_version == v and r.rtt != 0 and r.ip != ip:
+                        r.rtt = 0  # actually, it failed
+                        r.ip = f'Incorrect IP returned: {r.ip}'
+            raise DoneWithJobs(ip)
+            # logger.info(f"{fail_count} fails for {len(completed_jobs)} jobs")
+        top_ip_count = max(top_ip_count, ranked[0][1])
+        top_ip_target = max(top_ip_target, target)
+    if top_ip_count == 0:
+        are_we_there_yet.__dict__.pop('found_multiple_ips', None)  # show again if needed
+    return top_ip_count, top_ip_target
+
+
+def log_debug_details(logger, top_ip_count, pending_jobs_count, top_ip_bonus, top_ip_target):
+    if logger.isEnabledFor(logging.DEBUG):
+        try:
+            last_debug_msg = main_loop.last_debug_msg
+        except AttributeError:
+            last_debug_msg = ""
+        msg = f"have {top_ip_count} IPs plus {pending_jobs_count} pending,"
+        msg += f" want {top_ip_bonus}, need {top_ip_target}"
+        if msg != last_debug_msg:  # avoid repeating the same message
+            logger.debug(msg)
+        main_loop.last_debug_msg = msg
 
 
 class DoneWithJobs(Exception):
@@ -300,7 +363,6 @@ class DoneWithJobs(Exception):
 
 async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
     result = ""
-    error = None
     engine = init_db(args.dbfile)
     with Session(engine) as db_session:
         scores = dict()
@@ -325,54 +387,34 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
             limit=args.max_connections,
             family=ip_family,
         )
+        completed_jobs = list()
         async with aiohttp.ClientSession(
             connector=connector, timeout=http_timeout(args.timeout), trace_configs=[http_timer()]
         ) as aio_session:
             try:
                 logger.info(f"requests (need {args.minimum_match} matches):")
                 pending_jobs_count = 0
-                ip_counts = dict()
-                ip_counts[4] = dict()  # number of occurrences for each received IPv4
-                ip_counts[6] = dict()  # number of occurrences for each received IPv6
-                fail_count = 0
                 while True:  # spawn and collect jobs
-                    max_ipv0_count = 0  # max of IPv4 and IPv6
-                    for v in (4, 6):
-                        max_ip_count = 0 if len(ip_counts[v]) == 0 else max(ip_counts[v].values())
-                        if len(ip_counts[v]) > 1:
-                            error = f"multiple IPs received: {ip_counts[v]}"
-                            raise DoneWithJobs
-                        if max_ip_count >= args.minimum_match:  # we have sufficient results
-                            ip = max(ip_counts[v], key=ip_counts[v].get)
-                            msg = f"IP found: {ip} ({ip_counts[v][ip]} successes,"
-                            logger.info(f"{msg} {fail_count} failures)")
-                            result = ip
-                            raise DoneWithJobs
-                        max_ipv0_count = max(max_ipv0_count, max_ip_count)
-                    # spawn another get_ip() job if needed
-                    wanted_count = args.minimum_match + args.overkill
-                    jobs_count = max_ipv0_count + pending_jobs_count
-                    if logger.isEnabledFor(logging.DEBUG):
-                        try:
-                            last_debug_msg = main_loop.last_debug_msg
-                        except AttributeError:
-                            last_debug_msg = ""
-                        msg = f"have {max_ipv0_count} IPs plus {pending_jobs_count} pending,"
-                        msg += f" want {wanted_count}, need {args.minimum_match}"
-                        if msg != last_debug_msg:  # avoid repeating the same message
-                            logger.debug(msg)
-                        main_loop.last_debug_msg = msg
-                    if args.overkill > 0 and wanted_count > jobs_count and len(scores) == 0:
-                        msg = f"not enough providers for {wanted_count} requests"
+                    # calculate what is still needed
+                    top_ip_count, top_ip_target = are_we_there_yet(completed_jobs, args, logger)
+                    top_ip_bonus = top_ip_target + args.overkill
+                    jobs_count = top_ip_count + pending_jobs_count
+                    # logging, overkill check
+                    log_debug_details(
+                        logger, top_ip_count, pending_jobs_count, top_ip_bonus, top_ip_target
+                    )
+                    if args.overkill > 0 and top_ip_bonus > jobs_count and len(scores) == 0:
+                        msg = f"not enough providers for {top_ip_bonus} requests"
                         logger.warning(f"{msg}; ignoring '--overkill'")
                         args.overkill = 0
-                        wanted_count = args.minimum_match + args.overkill
-                    if wanted_count > jobs_count and pending_jobs_count < args.max_connections:
+                        top_ip_bonus = top_ip_target + args.overkill
+                    # spawn another get_ip() job if needed
+                    if top_ip_bonus > jobs_count and pending_jobs_count < args.max_connections:
                         # TCPConnector(limit=...) does this too, but checking args.max_connections
                         # ... here delays "not enough providers" so we can collect more responses
                         if len(scores) == 0:
                             error = f"not enough providers for {args.minimum_match} matches"
-                            raise DoneWithJobs
+                            raise NetworkError(error)
                         p_id = weighted_random(scores)
                         statement = select(Parrot).where(Parrot.id == p_id)
                         p = db_session.exec(statement).one_or_none()
@@ -380,42 +422,41 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
                         pending_jobs_count += 1
                         logger.debug(f"launching task {p.id}")
                         p.del_keys_of_same_parrot(scores, logger)  # don't use this parrot again
+                        main_loop.last_debug_msg = ""  # repeat debug msg after launch even if same
                     await asyncio.sleep(0.0001)
                     # process completed jobs
                     try:
                         r: FetchedIP = q.get_nowait()  # may raise QueueEmpty
-                        r.parrot.attempt_count += 1
-                        if r.rtt != 0:  # got valid IP of some sort
-                            if r.ip != '':
-                                v = r.ip_version
-                                ip_counts[v][r.ip] = ip_counts[v].get(r.ip, 0) + 1
-                            r.parrot.success_count += 1
-                            r.parrot.total_rtt += r.rtt
-                            portion = f"{r.parrot.success_count} of {r.parrot.attempt_count}"
+                        completed_jobs.append(r)  # keep a list to write to DB later
+                        if r.rtt != 0:  # got valid IP
+                            # note success count will be wrong if IP is later found to be wrong
+                            p = f"{r.parrot.success_count + 1} of {r.parrot.attempt_count + 1}"
                             msg = f"    {r.parrot.url()} → {r.ip} ({r.rtt} ms;"
-                            logger.info(f"{msg} {portion} succeeded)")
+                            logger.info(f"{msg} {p} succeeded)")
                         else:
-                            r.parrot.last_errmsg = r.ip
-                            portion = f"{r.parrot.success_count} of {r.parrot.attempt_count}"
+                            p = f"{r.parrot.success_count} of {r.parrot.attempt_count + 1}"
                             msg = f"    {r.parrot.url()} → {r.ip[:40]}"
-                            logger.info(f"{msg} ({portion} succeeded)")
-                            fail_count += 1
+                            logger.info(f"{msg} ({p} succeeded)")
                         pending_jobs_count -= 1
-                        db_session.add(r.parrot)
-                        db_session.commit()
-                        if fail_count >= args.max_failures:
-                            error = f"{fail_count} requests failed; giving up"
-                            raise DoneWithJobs
                     except asyncio.QueueEmpty:
                         await asyncio.sleep(0.005)  # reduce looping when waiting on responses
                     await asyncio.sleep(0.0001)
-            except DoneWithJobs:
-                pass
-        connector.close()
-    if error is not None:
-        logger.error(error)
-        if args.exception_level >= 1:
-            raise NetworkError(error)
+            except DoneWithJobs as e:
+                result = str(e)
+            except NetworkError as e:
+                logger.error(e)
+                if args.exception_level >= 1:
+                    raise NetworkError(e)
+        await connector.close()
+        for r in completed_jobs:
+            r.parrot.attempt_count += 1
+            if r.rtt != 0:  # got valid IP of some sort
+                r.parrot.success_count += 1
+                r.parrot.total_rtt += r.rtt
+            else:
+                r.parrot.last_errmsg = r.ip
+            db_session.add(r.parrot)
+            db_session.commit()
     return result
 
 
@@ -437,10 +478,9 @@ def db_pathname(create_dir=False):
     return os.path.join(config_dir, f'data.sqlite')
 
 
-def init_db(db_file=''):
+def init_db(db_file):
     if hasattr(init_db, 'engine'):  # only need to initialize once
         return init_db.engine
-    GrepIPs.grep_ips_test()
     warnings.filterwarnings(  # https://github.com/tiangolo/sqlmodel/issues/189#issuecomment-1018014753
         "ignore", ".*Class SelectOfScalar will not make use of SQL compilation caching.*"
     )
